from __future__ import division

import re
import random
import math

import pygame

import data
from helper import OrderedRenderUpdates


class Thing(pygame.sprite.Sprite):
    """
    A basic object which is subject to rules.
    """
    what = 'Thingy'
    group = None
    collidables = {}

    def __init__(self, images, pos, speed=0, world=None):
        pygame.sprite.Sprite.__init__(self)

        self._image_paths = images
        self._images = []
        for image in images:
            image = pygame.image.load(data.load(image)).convert_alpha()
            self._images.append(image)
        self.image = self._images[0]

        screen = pygame.display.get_surface()
        self.area = screen.get_rect()
        self.rect = self.image.get_rect()
        self.rect.x, self.rect.y = pos
        
        self.speed = speed
        self._old_direction = self.direction = (0, 0)
        self.world = world

        # Queues for effects resulting from rules.
        self._init_queue = []
        self._finish_queue = []

    def copy(self):
        return self.__class__(self._image_paths, (self.rect.x, self.rect.y), self.speed, self.world)
    
    def _init(self):
        while self._init_queue:
            f = self._init_queue.pop()
            f(self)

    def _finish(self):
        while self._finish_queue:
            f = self._finish_queue.pop()
            f(self)

    def update(self, groups=None):
        self._move(groups)

    def _move(self, groups=None):
        groups = groups or {}

        speed = self.speed

        # Check "soft" collision to adjust speed.
        for name, modifier in self.collidables.items():
            if name in groups:
                if pygame.sprite.spritecollide(self, groups[name], 0):
                    speed *= modifier
                    if name == 'ICE':
                        self.direction = self._old_direction

        dx = self.direction[0] * speed
        dy = self.direction[1] * speed
        newpos = self.rect.move((dx, dy))
        if self.area.contains(newpos):
            self.rect = newpos

            # Check "hard" collision.
            for name, modifier in self.collidables.items():
                if name in groups:
                    if pygame.sprite.spritecollide(self, groups[name], 0):
                        if modifier == 0:
                            # Go back.
                            self.rect.move_ip((-dx, -dy))
                            break

        # Reset direction (it needs to be set every frame).
        self._old_direction = self.direction
        self.direction = (0, 0)

        # Update image.
        if len(self._images) == 8:
            if dx > 0:
                self.image = self._images[0]
                self._images[0:2] = self._images[1::-1]
            elif dx < 0:
                self.image = self._images[2]
                self._images[2:4] = self._images[3:1:-1]
            elif dy < 0:
                self.image = self._images[4]
                self._images[4:6] = self._images[5:3:-1]
            elif dy > 0:
                self.image = self._images[6]
                self._images[6:8] = self._images[7:5:-1]


class Ball(Thing):
    group = what = 'BALL'
    group = 'BALL'
    collidables = {'TREE': 0, 'SAND': 0.5, 'ICE': 2}


class Tree(Thing):
    what = '(TREE|VEGETATION)'
    group = 'TREE'
    collidables = {'SAND': 0.5, 'BALL': 0, 'WATER': 0, 'ICE': 2, 'PLAYER': 0}


class Sand(Thing):
    group = what = 'SAND'
    collidables = {}


class Water(Thing):
    what = '(WATER|LAKE|POND)'
    group = 'WATER'
    collidables = {}


class Ice(Water):
    group = what = 'ICE'


class FreeWilling(Thing):
    what = 'Free willing'
    group = None
    collidables = {'TREE': 0, 'SAND': 0.5, 'WATER': 0, 'ICE': 2}


class Cat(FreeWilling):
    what = '(CAT|ANIMAL)'
    group = 'CAT'

    def __init__(self, *args, **kwargs):
        FreeWilling.__init__(self, *args, **kwargs)
        self.dr = []

    def get_direction(self):
        # Calculate a random direction.
        x0, y0 = (random.choice([-1,0,1]), random.choice([-1,0,1]))
        
        # Calculate direction towards player.
        dists = []
        for object in [o for o in self.world if o.alive]:
            if (re.match(object.what, 'PLAYER') or
                    re.match(object.what, 'CAT')):
                tx, ty = object.rect.x, object.rect.y
                dx = tx - self.rect.x
                dy = ty - self.rect.y
                dr = math.sqrt(dx**2 + dy**2)
                dists.append((dr, dx, dy))
        if not dists: return (0, 0)

        # Run for closest player.
        dists.sort()
        dr, dx, dy = dists[0]

        if dx != 0:
            x1 = dx/abs(dx)
        else:
            x1 = 0
        if dy != 0:
            y1 = dy/abs(dy)
        else:
            y1 = 0

        # Now we return a random direction or the direction
        # to the ball, proportional to how well the strategy
        # is working. If the distance is not getting smaller,
        # we favor the random direction.
        if self.dr:
            if len(self.dr) > 15: self.dr.pop(0)
            r = (sum(self.dr)/15.0 - dr)/3.5
            r = max(0, r)
            r = min(1, r)
        else:
            r = 0
        self.dr.append(dr)
        x = r*x1 + (1-r)*x0
        y = r*y1 + (1-r)*y0
        return (x, y)

    def set_direction(self, direction):
        pass

    direction = property(get_direction, set_direction)


class Dog(Cat):
    what = '(DOG|ANIMAL)'
    group = 'DOG'

    def _move(self, groups=None):
        FreeWilling._move(self, groups)

        # Check collision with player and kill him!
        pygame.sprite.spritecollide(self, groups['PLAYER'], 1)


class Person(FreeWilling):
    what = '(PLAYER|PERSON|PEOPLE)'
    group = 'PLAYER'
    alive = True

    def kill(self, *args, **kwargs):
        self.alive = False
        FreeWilling.kill(self, *args, **kwargs)


class Player(Person):
    name = 'Galvin'


class Computer(Person):
    name = 'Suzanne'

    def __init__(self, *args, **kwargs):
        FreeWilling.__init__(self, *args, **kwargs)
        self.dr = []

    def get_direction(self):
        # Calculate a random direction.
        x0, y0 = (random.choice([-1,0,1]), random.choice([-1,0,1]))
        
        # Calculate direction towards ball.
        for object in self.world:
            if re.match(object.what, 'BALL'):
                tx, ty = object.rect.x, object.rect.y
                dx = tx - self.rect.x
                dy = ty - self.rect.y
                dr = math.sqrt(dx**2 + dy**2)
                if dx != 0:
                    x1 = dx/abs(dx)
                else:
                    x1 = 0
                if dy != 0:
                    y1 = dy/abs(dy)
                else:
                    y1 = 0
                break
        else:
            dr = x1 = y1 = 0

        # Now we return a random direction or the direction
        # to the ball, proportional to how well the strategy
        # is working. If the distance is not getting smaller,
        # we favor the random direction.
        if self.dr:
            if len(self.dr) > 15: self.dr.pop(0)
            r = (sum(self.dr)/15.0 - dr)/3.5
            r = max(0, r)
            r = min(1, r)
        else:
            r = 0
        self.dr.append(dr)
        x = r*x1 + (1-r)*x0
        y = r*y1 + (1-r)*y0
        return (x, y)

    def set_direction(self, direction):
        pass

    direction = property(get_direction, set_direction)


def create_world(level=None):
    objects = []
    groups = {}
    allsprites = OrderedRenderUpdates()
    groups['ALL'] = allsprites

    # Add player.
    player = Player(
            ["player_e1.png",
             "player_e2.png",
             "player_w1.png",
             "player_w2.png",
             "player_n1.png",
             "player_n2.png",
             "player_s1.png",
             "player_s2.png"],
            (0, 480-32), speed=4)
    player.name = 'Roberto'
    objects.append(player)

    computer = Computer(
            ["computer_e1.png",
             "computer_e2.png",
             "computer_w1.png",
             "computer_w2.png",
             "computer_n1.png",
             "computer_n2.png",
             "computer_s1.png",
             "computer_s2.png"],
             (0, 0), speed=3, world=objects)
    objects.append(computer)

    groups['PLAYER'] = pygame.sprite.Group((player, computer))

    if level is not None:
        f = data.load("level%d.dat" % level)
        y = 0
        sprites = []
        for line in f:
            x = 0
            if line.startswith('#'): continue

            for char in line:
                if char == 'C':
                    computer.rect.x = x
                    computer.rect.y = y
                elif char == 'P':
                    player.rect.x = x
                    player.rect.y = y
                elif char == 'S':
                    sand = Sand(["sand.png"], (x, y))
                    group = groups.setdefault('SAND', pygame.sprite.Group())
                    group.add(sand)
                    sprites.append((0.0, sand))
                    objects.append(sand)
                elif char == 'T':
                    tree = Tree(["tree.png"], (x, y))
                    group = groups.setdefault('TREE', pygame.sprite.Group())
                    group.add(tree)
                    sprites.append((1.0, tree))
                    objects.append(tree)
                elif char == 'W':
                    water = Water(["water.png"], (x, y))
                    group = groups.setdefault('WATER', pygame.sprite.Group())
                    group.add(water)
                    sprites.append((0.5, water))
                    objects.append(water)
                elif char == 'B':
                    ball = Ball(["ball.png"], (x, y))
                    group = groups.setdefault('BALL', pygame.sprite.Group())
                    group.add(ball)
                    sprites.append((2.0, ball))
                    objects.append(ball)
                elif char == 'D':
                    dog = Dog(
                            ["dog_e1.png",
                             "dog_e2.png",
                             "dog_w1.png",
                             "dog_w2.png",
                             "dog_n1.png",
                             "dog_n2.png",
                             "dog_s1.png",
                             "dog_s2.png"],
                             (x, y), speed=2, world=objects)
                    sprites.append((3.0, dog))
                    objects.append(dog)
                elif char == 'A':
                    cat = Cat(
                            ["cat_e1.png",
                             "cat_e2.png",
                             "cat_w1.png",
                             "cat_w2.png",
                             "cat_n1.png",
                             "cat_n2.png",
                             "cat_s1.png",
                             "cat_s2.png"],
                             (x, y), speed=2, world=objects)
                    sprites.append((2.9, cat))
                    objects.append(cat)
                elif char == 'I':
                    ice = Ice(["ice.png"], (x, y))
                    group = groups.setdefault('ICE', pygame.sprite.Group())
                    group.add(ice)
                    sprites.append((0.75, ice))
                    objects.append(ice)

                x += 32
            y += 32

        # Add sprites in proper order.
        sprites.sort()
        for w, sprite in sprites:
            allsprites.add(sprite)

    else:
        # Add sand.     
        n = random.expovariate(0.08)
        groups['SAND'] = sandGroup = pygame.sprite.Group()
        while len(sandGroup) < n:
            x = random.randint(0, 640-32)
            y = random.randint(32, 480-64)
            sand = Sand(["sand.png"], (x, y))
            if not pygame.sprite.spritecollide(sand, allsprites, 0):
                sandGroup.add(sand)
                allsprites.add(sand)
                objects.append(sand)
                
        # Add some random trees.
        n = random.expovariate(0.08)
        groups['TREE'] = treeGroup = pygame.sprite.Group()
        while len(treeGroup) < n:
            x = random.randint(0, 640-32)
            y = random.randint(32, 480-64) # leave a starting place for players
            tree = Tree(["tree.png"], (x, y))
            if not pygame.sprite.spritecollide(tree, allsprites, 0):
                treeGroup.add(tree)
                allsprites.add(tree)
                objects.append(tree)

        # Add ball.
        groups['BALL'] = ballGroup = pygame.sprite.Group()
        while len(ballGroup) < 1:
            x = random.randint(0, 640-32)
            y = random.randint(32, 480-64) # leave a starting place for players
            ball = Ball(["ball.png"], (x, y))
            if not pygame.sprite.spritecollide(ball, allsprites, 0):
                ballGroup.add(ball)
                allsprites.add(ball)
                objects.append(ball)

    allsprites.add(computer)
    allsprites.add(player)

    return player, objects, groups, allsprites
